import pytest
import pandas as pd
from opendc_eemm.preprocess import aggregate_predictions
from opendc_eemm.visualization import plot_power_draw

@pytest.mark.parametrize('value, expected', [(1, 1)])
def test_test(value, expected):
    assert value == expected

def test_aggregate_predictions():
    with pytest.raises(ValueError):
        aggregate_predictions(pd.DataFrame(), 'max')

def test_plot_power_draw():
    test_dict = {'governor': 'superscaler'}
    with pytest.raises(RuntimeError):
        plot_power_draw(pd.DataFrame([test_dict]), 'powersave', 0)



import numpy as np
import pandas as pd
import os
from tqdm import tqdm

from .preprocess import watt_to_mwh
from .market import extend_spot_prices


def compute_agreement_score(df_pred, df_da, df_im):
    epsilon = 10
    df_da = extend_spot_prices(df_da, df_pred)
    perf = (
        (
            (
                np.sign(df_im['shortage-price'])
                == np.sign(df_pred['shortage-predicted'])
            ) & (
                np.sign(
                    round(
                        df_im['shortage-price'] - df_da['dayahead-price'],
                        epsilon
                    )
                ) == np.sign(
                    round(
                        df_pred['shortage-predicted'] - df_da['dayahead-price'],
                        epsilon
                    )
                )
            )
        ).value_counts().to_dict()
    )
    perf[False] = perf[False] if False in perf else 0

    return perf[True] / (perf[False] + perf[True]) * 100


def process_traces_for_scheduling(df_trace):
    df = (
        df_trace[df_trace['power-model'] != 'BASE'].drop(
            [
                'duration', 'state', 'vm-count', 'requested-burst',
                'granted-burst', 'interfered-burst', 'cpu-usage', 'cpu-demand',
                'cores'
            ],
            axis=1
        )
        # Average two power models
        .groupby(['governor', 'host-id', 'timestamp']).mean()
        # Aggregate hosts' data
        .groupby(['governor', 'timestamp']).sum().reset_index()
        # Aggregate to ISP (15min)
        .groupby(['governor',
                  pd.Grouper(freq='15min',
                             key='timestamp')]).sum().drop('no-dvfs')
    )
    df['consumption'] = watt_to_mwh(df['power-draw'])
    return df


def schedule(damping, df_pred, df_da, df_trace, save_path=None):
    df_da = extend_spot_prices(df_da, df_pred)
    # Initialize an empty schedule.
    sched = []
    governor = None
    timestamp = None
    total_oc = 0
    prev_overcommission = 0
    oc_counter = 0

    for i, inference in tqdm(df_pred.iterrows(), total=len(df_pred)):
        spot_price = df_da.iloc[i]['dayahead-price']
        # Manually take the mean for performance reasons.
        curr_overcommission = total_oc / (i + 1)

        if curr_overcommission > prev_overcommission:
            # OC level increased
            oc_counter += 1
        else:
            # OC level decreased.
            oc_counter = 0

        prev_overcommission = curr_overcommission

        if inference['shortage-predicted'] <= 0:
            #             if oc_counter >= damping:
            #                 governor = 'ondemand'
            #                 oc_counter = 0
            #             else:
            governor = 'performance'
        else:
            if inference['shortage-predicted'] > spot_price:
                governor = 'powersave'
            else:
                if oc_counter >= damping:
                    governor = 'conservative'
                    oc_counter = 0
                else:
                    governor = 'ondemand'

        record = df_trace.loc[governor].reset_index().iloc[i % len(df_trace)]
        total_oc += record['overcommissioned-burst']
        sched.append(
            {
                'timestamp': inference['interval-start'],
                'governor': governor
            }
        )

    df_sched = pd.DataFrame(sched)
    if save_path and not os.path.exists(save_path):
        os.makedirs(save_path)
        df_sched.to_csv(f"{save_path}/dvfs_schedule.csv", index=False)

    return df_sched


def compute_energy_cost(df_sched, df_price):
    imbalance = df_sched.consumption - df_sched.baseload
    im_cost = 0

    for i, ledger in tqdm(df_price.iterrows(), total=len(df_price)):

        im_consumption = imbalance.iloc[i]

        if im_consumption > 0:
            # Shortage #
            im_cost += (im_consumption * ledger['shortage-price'])  # BRP -> TSO
        elif im_consumption < 0:
            # Surplus #
            im_cost += (
                im_consumption * ledger['spot-price']
            )  # TSO -> BRP (two-price system)
        else:
            # Spot on #
            pass

    total_cost = im_cost + (df_sched.baseload * df_price['spot-price']).sum()
    return total_cost


def get_governor_performance(governor, df_dc, df_price):
    # Extend simulation results.
    repeat = int(np.ceil(len(df_price) / len(df_dc.loc[governor])))
    df_governor = pd.concat(
        [df_dc.loc[governor].reset_index()] * repeat, ignore_index=True
    ).iloc[:len(df_price)]

    return df_governor, {
        'governor': governor,
        'cost': compute_energy_cost(df_governor, df_price),
        'consumption': df_governor.consumption.sum(),
        'overcommission': df_governor['overcommissioned-burst'].sum()
    }


def get_schedule_performance(df_sched, df_price, governor_perf):
    consumption = df_sched.consumption.sum()
    oc = df_sched['overcommissioned-burst'].sum()
    cost = compute_energy_cost(df_sched, df_price)

    return {
        'against':
            governor_perf['governor'],
        'cost %': (cost - governor_perf['cost']) / governor_perf['cost'] * 100,
        'energy %':
            (consumption - governor_perf['consumption']) /
            governor_perf['consumption'] * 100,
        'overcommission %':
            (oc - governor_perf['overcommission']) /
            governor_perf['overcommission'] * 100,
    }



from itertools import cycle
import pandas as pd
import numpy as np

from .preprocess import aggregate_consumption

extend_spot_prices = (
    lambda df_da, df_pred: pd.concat(
        [df_da] * (len(df_pred) // len(df_da)), ignore_index=True
    ).sort_values(by='interval-start').reset_index(drop=True)
)


def compute_agreement_score(df_pred, df_da, df_im):
    epsilon = 10
    df_da = extend_spot_prices(df_da, df_pred)
    perf = (
        (
            (
                np.sign(df_im['shortage-price'])
                == np.sign(df_pred['shortage-predicted'])
            ) & (
                np.sign(
                    round(
                        df_im['shortage-price'] - df_da['spot-price'], epsilon
                    )
                ) == np.sign(
                    round(
                        df_pred['shortage-predicted'] - df_da['spot-price'],
                        epsilon
                    )
                )
            )
        ).value_counts().to_dict()
    )

    perf[False] = perf[False] if False in perf else 0

    return perf[True] / (perf[False] + perf[True]) * 100


def garner_costs(od_price, df_trace, df_da, df_im):
    df_linear = aggregate_consumption(
        df_trace, power_model='LINEAR', freq='15min', preprocess=True
    )
    df_sqrt = aggregate_consumption(
        df_trace, power_model='SQRT', freq='15min', preprocess=True
    )
    df_base = aggregate_consumption(
        df_trace, power_model='BASE', freq='15min', preprocess=True
    )

    im_cost = {
        'baseload-LINEAR':
            (df_base.consumption * df_im['shortage-price']).sum(),
        'baseload-SQRT': (df_base.consumption * df_im['shortage-price']).sum(),
        'peakload-LINEAR':
            (
                (df_linear.consumption - df_base.consumption) *
                df_im['shortage-price']
            ).sum(),
        'peakload-SQRT':
            (
                (df_sqrt.consumption - df_base.consumption) *
                df_im['shortage-price']
            ).sum(),
        'fullload-LINEAR':
            (df_linear.consumption * df_im['shortage-price']).sum(),
        'fullload-SQRT': (df_sqrt.consumption * df_im['shortage-price']).sum(),
    }
    df_im_costs = pd.DataFrame(im_cost.items(), columns=['load', 'price'])
    df_im_costs['market'] = 'imbalance'
    df_im_costs['model'] = ''

    for i, row in df_im_costs.iterrows():
        tmp = row['load']
        df_im_costs.at[i, 'load'] = tmp.split('-')[0]
        df_im_costs.at[i, 'model'] = tmp.split('-')[1]

    df_base_da = aggregate_consumption(df_base, '1h')
    df_linear_da = aggregate_consumption(df_linear, '1h')
    df_sqrt_da = aggregate_consumption(df_sqrt, '1h')

    da_costs = {
        'baseload-LINEAR':
            (df_base_da.consumption * df_da['dayahead-price']).sum(),
        'baseload-SQRT':
            (df_base_da.consumption * df_da['dayahead-price']).sum(),
        'peakload-LINEAR':
            (
                (df_linear_da.consumption - df_base_da.consumption) *
                df_da['dayahead-price']
            ).sum(),
        'peakload-SQRT':
            (
                (df_sqrt_da.consumption - df_base_da.consumption) *
                df_da['dayahead-price']
            ).sum(),
        'fullload-LINEAR':
            (df_linear_da.consumption * df_da['dayahead-price']).sum(),
        'fullload-SQRT':
            (df_sqrt_da.consumption * df_da['dayahead-price']).sum(),
    }
    df_da_costs = pd.DataFrame(da_costs.items(), columns=['load', 'price'])
    df_da_costs['market'] = 'day-ahead'
    df_da_costs['model'] = ''

    for i, row in df_da_costs.iterrows():
        tmp = row['load']
        df_da_costs.at[i, 'load'] = tmp.split('-')[0]
        df_da_costs.at[i, 'model'] = tmp.split('-')[1]
    df_da_costs

    od_costs = {
        'baseload-LINEAR': (df_base.consumption * od_price).sum(),
        'baseload-SQRT': (df_base.consumption * od_price).sum(),
        'peakload-LINEAR':
            ((df_linear.consumption - df_base.consumption) * od_price).sum(),
        'peakload-SQRT':
            ((df_sqrt.consumption - df_base.consumption) * od_price).sum(),
        'fullload-LINEAR': (df_linear.consumption * od_price).sum(),
        'fullload-SQRT': (df_sqrt.consumption * od_price).sum(),
    }

    df_od_costs = pd.DataFrame(od_costs.items(), columns=['load', 'price'])
    df_od_costs['market'] = 'on-demand'
    df_od_costs['model'] = ''

    for i, row in df_od_costs.iterrows():
        tmp = row['load']
        df_od_costs.at[i, 'load'] = tmp.split('-')[0]
        df_od_costs.at[i, 'model'] = tmp.split('-')[1]

    return pd.concat([df_da_costs, df_im_costs, df_od_costs])


def compute_imbalance_cost(
    df_daload, df_fullload, df_price, is_one_price=False
):
    df_imload = df_daload.copy()
    df_imload['consumption'] = df_fullload.consumption - df_daload.consumption
    schedule = cycle(df_imload.consumption)
    im_cost = 0

    not_consider_state = True  # Shortcircuit state conditions.

    for _, isp in df_price.iterrows():
        im_consumption = next(schedule)

        if not_consider_state or isp.state >= 0:
            ''' Under production (Upwards or No regulation) '''
            if im_consumption > 0:
                # Shortage #
                im_cost += (
                    im_consumption * isp['shortage-price']
                )  # BRP -> TSO
            elif im_consumption < 0:
                # Surplus #
                #               # TSO -> BRP (one-price/two-price system)
                sell = (im_consumption * isp['surplus-price']) if is_one_price \
                        else (im_consumption * isp['spot-price']) # Two-price system
                im_cost += sell
            else:
                # Spot on #
                pass

        else:
            ''' Over production (Downwards regulation) '''
            if im_consumption > 0:
                # Shortage #
                #                 im_cost -= (im_consumption * isp['shortage-price']) # TSO -> BRP (one-price)
                im_cost -= (
                    im_consumption * isp['spot-price']
                )  # TSO -> BRP (two-price)
            elif im_consumption < 0:
                # Surplus #
                im_cost -= (im_consumption * isp['surplus-price'])  # BRP -> TSO
            else:
                # Spot on #
                pass

    return im_cost



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.dates import DateFormatter
import datetime
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_percentage_error
import itertools
from itertools import cycle
from statannot import add_stat_annotation
from matplotlib.patches import *
import matplotlib.dates as mdates
from datetime import timedelta

import matplotlib.patches as mpatches
from matplotlib.patches import Rectangle
from matplotlib.lines import Line2D
from matplotlib import rc
import matplotlib.font_manager

plt.style.use('default')
sns.set_style("whitegrid", {'grid.linestyle': '--'})

from .preprocess import *


def set_hatches(ax, cycle):
    hatches = itertools.cycle(
        ['/', '\\', 'x', '-', '.', '//', '+', '*', 'o', 'O']
    )
    for i, bar in enumerate(ax.patches):
        if i % cycle == 0:
            hatch = next(hatches)
        bar.set_hatch(hatch)


def show_values_on_bars(
    axs,
    h_v="v",
    pad=0.4,
    append='',
):
    def _show_on_single_plot(ax):
        if h_v == "v":
            ''' Vertical plot '''
            for p in ax.patches:
                _x = p.get_x() + p.get_width() / 2
                _y = p.get_y() + p.get_height() + float(pad)
                value = float(p.get_height())
                ax.text(_x, _y, f"{value:.3f}{append}", ha="center")
        elif h_v == "h":
            ''' Horizontal plot '''
            for p in ax.patches:
                _x = p.get_x() + p.get_width() + float(pad)
                _y = p.get_y() + p.get_height() / 1.5
                value = float(p.get_width())
                ax.text(_x, _y, f"{value:.3f}{append}", ha="left")

    if isinstance(axs, np.ndarray):
        for idx, ax in np.ndenumerate(axs):
            _show_on_single_plot(ax)
    else:
        _show_on_single_plot(axs)


def set_sns_legend(ax, new_loc='best', title=None, **kws):
    old_legend = ax.legend_
    handles = old_legend.legendHandles
    labels = [t.get_text() for t in old_legend.get_texts()]
    title = old_legend.get_title().get_text() if title is None else title
    ax.legend(handles, labels, loc=new_loc, title=title, **kws)


def plot_power_draw(df, governor, core_freq):
    governors = set(df.governor)
    if governor not in governors:
        raise RuntimeError(f"{governor=} must be in {governors}")

    df = prepare_traces_for_plot(df, core_freq)
    hue_order = ['SQRT', 'LINEAR']
    ax = sns.lineplot(
        data=df[(df.governor == governor) & (df['power-model'] != 'BASE')],
        hue_order=hue_order,
        x="timestamp",
        y='power-draw',
        hue='power-model',
        style='power-model',
        ci=90,
        n_boot=1500
    )

    date_form = DateFormatter("%d %b")
    ax.xaxis.set_major_formatter(date_form)
    ax.set_title(f"Governor: {governor}")
    ax.legend(title='Power Model', fontsize=10, loc='best')
    ax.set(ylabel='Instant Power Draw [Watt]', xlabel=None)
    plt.xticks(rotation=30)
    plt.show()


def plot_overcommission(df, core_freq):
    df = prepare_traces_for_plot(df, core_freq)

    ax = sns.lineplot(
        data=df,
        x="timestamp",
        y='overcommissioned-burst',
        hue='governor',
        style='governor',
        ci=90,
        n_boot=1500
    )

    date_form = DateFormatter("%d\ %b")
    ax.xaxis.set_major_formatter(date_form)
    ax.set(xlabel=None, ylabel='Overcommissioned CPU Cycle')
    ax.legend(title='Governor', fontsize=10, loc='best')
    sns.despine(left=True)
    plt.xticks(rotation=30)
    plt.show()


def plot_market_costs(df):
    ax = sns.barplot(
        data=df,
        y='load',
        x='price',
        hue='market',
        palette='pastel',
        estimator=np.mean,
        ec='k',
        errwidth=2,
        capsize=0.1,
        errcolor='k'
    )
    ax.set(ylabel='Load Type', xlabel='Energy Cost [€]')
    ax.legend(title='Market')
    sns.despine(left=True)
    set_hatches(ax, 3)
    show_values_on_bars(ax, 'h', 100, ' €')
    plt.show()


def plot_purchase_comparison(df_costs):
    da_base = df_costs[(df_costs.load == 'baseload') &
                       (df_costs.market == 'day-ahead')]['price'].mean()
    im_peak = df_costs[(df_costs.load == 'peakload') &
                       (df_costs.market == 'imbalance')]['price'].mean()
    da_im = da_base + im_peak

    da_full = df_costs[(df_costs.load == 'fullload') &
                       (df_costs.market == 'day-ahead')]['price'].mean()
    od_full = df_costs[(df_costs.load == 'fullload') &
                       (df_costs.market == 'on-demand')]['price'].mean()

    odda_reduction = (od_full - da_full) / od_full * 100
    oddaim_reduction = (od_full - da_im) / od_full * 100
    daim_reduction = (da_full - da_im) / da_full * 100

    od_cat = 'On-Demand (full load)'
    da_cat = 'Day-Aead (full load)'
    da_im_cat = 'Day-Aead (base load)\n+ Balancing (peak load)'

    x = [od_cat, da_cat, da_im_cat]
    y = [od_full, da_full, da_im]

    order = x
    ax = sns.barplot(
        x=x, y=y, order=order, palette='deep', ec='k', lw=2, fill=True
    )
    ax.set_ylabel('Energy Cost [€]')

    set_hatches(ax, 1)

    show_values_on_bars(ax, 'v', 100, ' €')

    add_stat_annotation(
        ax,
        x=x,
        y=y,
        order=order,
        box_pairs=[(da_cat, da_im_cat), (od_cat, da_cat), (od_cat, da_im_cat)],
        text_annot_custom=[
            f"{abs(daim_reduction): 0.1f}\%", f"{odda_reduction: 0.1f}\%",
            f"{oddaim_reduction: 0.1f}\%"
        ],
        pvalues=[0, 0, 0],
        perform_stat_test=False,
        line_offset_to_box=0.08,
        text_offset=0.08,
        text_format='full',
        loc='inside',
        verbose=0
    )
    sns.despine(left=True)
    plt.show()



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_percentage_error
import itertools
from itertools import cycle

watt_to_mwh = lambda x: np.float128(x * 5 * 60 * 2.7777777777778e-10)


def preprocess_dayahead(df):
    df['interval-start'] = pd.to_datetime(
        df['MTU (CET)'].apply(lambda cell: cell.split('-')[0].strip()),
        format='%d.%m.%Y %H:%M'
    )
    df = df.drop([
        'MTU (CET)', 'BZN|NL'
    ], axis=1).set_index('interval-start').reset_index()  # Reorder columns.
    df.columns = ['interval-start', 'dayahead-price']
    return df


# Default frequency == one ISP (15min).
def aggregate_predictions(df, agg, freq='15min'):
    if agg not in ['first', 'last', 'mean']:
        raise ValueError(f"{agg=} is not one of ['first', 'last', 'mean'].")

    df.columns = df.columns.str.replace('_', '-')
    df['interval-start'] = pd.to_datetime(df['interval-start'])

    return df.groupby([pd.Grouper(freq=freq,
                                  key='interval-start')]).agg(agg).reset_index()


def preprocess_imbalance(df_im):
    df = pd.DataFrame()
    df['shortage-price'] = df_im.Consume
    df['surplus-price'] = df_im.Feed
    return df


def preprocess_traces(df, pue, map_governor=False):
    df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
    df.columns = df.columns.str.replace('_', '-')
    df['power-model'] = df['power-model'].apply(lambda s: s.replace('_', '-'))
    # Use PUE to compute the power draw of the whole datacenter
    df['power-draw'] = df['power-draw'] * pue
    if map_governor:
        df['governor'] = df['governor'].map({ # Map model names here.
            'ConservativeScalingGovernor[threshold=0.8,stepSize=-1.0]': 'conservative',
            'OnDemandScalingGovernor[threshold=0.8]': 'ondemand',
            'PerformanceScalingGovernor': 'performance',
            'PowerSaveScalingGovernor': 'powersave',
            'null': 'no-dvfs',
        })
    return df


def prepare_traces_for_plot(df, core_freq):
    usage_to_percent = lambda x: x['cpu-usage'] / (core_freq * x.cores) * 100
    demand_to_percent = lambda x: x['cpu-demand'] / (core_freq * x.cores) * 100

    df = df.groupby(
        [
            'governor', 'power-model', 'host-id',
            pd.Grouper(freq='1d', key='timestamp')
        ]
    ).mean().reset_index(level=['timestamp', 'host-id'])
    df['cpu-usage'] = df.apply(usage_to_percent, axis=1)
    df['cpu-demand'] = df.apply(demand_to_percent, axis=1)
    return df.reset_index()


def aggregate_consumption(
    df_raw, freq, kind='', power_model=None, preprocess=False
):
    df = pd.DataFrame()
    df_raw = df_raw[df_raw['power-model'] == power_model
                   ] if power_model else df_raw
    df['timestamp'] = pd.to_datetime(df_raw['timestamp'], unit='ms')

    if (preprocess):
        # Only keep the power draw for now.
        df_raw.columns = df_raw.columns.str.replace('_', '-')
        df['consumption'] = df_raw['power-draw']
        # * The original traces is of 5min-granularity.
        # * Jouls => MWh (1 J = 2.7777777777778 * 10^-10 MWh).
        df['consumption'] = watt_to_mwh(df.consumption)
    else:
        df['consumption'] = df_raw.consumption

    df = df.groupby([pd.Grouper(freq=freq, key='timestamp')]
                   ).sum().reset_index(level=['timestamp'])
    df['kind'] = f"{kind} {freq}"
    return df



"""Console script for opendc_eemm."""
from pprint import pprint
import pandas as pd
import argparse
import logging
import sys

from .__init__ import __version__
from .preprocess import *
from .market import *
from .visualization import *
from .decision import *
'''
Example commands:

trace:
    * opendc-eemm -t ./example/traces.parque --pue 1.58 trace -s power --g performance
    * opendc-eemm -t ./example/traces.parque --pue 1.58 trace -s oc

market:
    * opendc-eemm -t ./example/traces.parque market -s load -o 80.4 -d ./example/market/dayahead_2020.csv -i ./example/market/balancing_2020.csv
    * opendc-eemm -t ./example/traces.parque market -s strategy -o 80.4 -d ./example/market/dayahead_2020.csv -i ./example/market/balancing_2020.csv

decision:
    * opendc-eemm -t ./example/traces.parque decision -o score -f 12 -a first -d ./example/market/dayahead_2020.csv -i ./example/market/balancing_2020.csv -p ./example/fake_pred_sigma50.csv
    * opendc-eemm -t ./example/traces.parque decision -o schedule -f 12 -a first -d ./example/market/dayahead_2020.csv -i ./example/market/balancing_2020.csv -p ./example/fake_pred_sigma50.csv -s ./output
'''


def dispatch(args, df_trace):

    if args.cmd == 'trace':
        logging.info('Visualizing results.')
        if args.show == 'power':
            plot_power_draw(df_trace, args.governor, core_freq=args.frequency)
        elif args.show == 'oc':
            plot_overcommission(df_trace, core_freq=args.frequency)
    else:
        df_da = preprocess_dayahead(pd.read_csv(args.dayahead_prices))
        df_im = preprocess_imbalance(pd.read_csv(args.imbalance_prices))

        if args.cmd == 'market':
            df_costs = garner_costs(args.od_price, df_trace, df_da, df_im)
            logging.info('Visualizing results.')

            if args.show == 'load':
                plot_market_costs(df_costs)
            elif args.show == 'strategy':
                plot_purchase_comparison(df_costs)

        elif args.cmd == 'decision':
            df_pred = aggregate_predictions(
                pd.read_csv(args.predictions), args.aggregate
            )

            if args.option == 'score':
                score = compute_agreement_score(df_pred, df_da, df_im)
                logging.info(
                    f"The AA score of the predictions is {score: 0.3f}%."
                )
            elif args.option == 'schedule':
                logging.info('DVFS scheduler invoked.')
                df_power = process_traces_for_scheduling(df_trace)
                df_sched = schedule(
                    args.factor, df_pred, df_da, df_power, args.save_to
                )
                pprint(df_sched)

                if args.save_to:
                    logging.info(
                        f"`dvfs_schedule.csv` has been saved to {args.save_to}"
                    )
    return 0


def main():
    """Console script for opendc_eemm."""
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format='(%(asctime)s) OpenDC EEMM - [%(levelname)s] %(message)s'
    )

    parser = argparse.ArgumentParser(
        add_help=False,
        description="CLI of OpenDC Extension for Energy Modelling & Managament."
    )
    subparsers = parser.add_subparsers(
        dest='cmd', description='Available commands.'
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s ' + __version__,
        help="Show version number of the package and exit."
    )
    parser.add_argument(
        '-h',
        '--help',
        action='help',
        default=argparse.SUPPRESS,
        help='Show the help messages and exit.'
    )
    parser.add_argument(
        '-t',
        '--trace',
        required=True,
        metavar='path',
        help='Path to simulation results (expecting a Parque file).'
    )
    parser.add_argument(
        '--pue',
        default=1.58,
        type=float,
        metavar='float',
        help='PUE value of the simulatied datacenter.'
    )

    trace_parser = subparsers.add_parser(
        'trace', help='Visualize simulation results.'
    )
    trace_parser.add_argument(
        '-s',
        '--show',
        required=True,
        choices=['power', 'over-commission'],
        metavar="['power', 'oc']",
        help=
        "Choose 'power' to show power draw; choose 'oc' to show over-commissioned."
    )
    trace_parser.add_argument(
        '-f',
        '--frequency',
        type=float,
        default=2670,
        metavar='float',
        help='Frequency of simulated machines.'
    )
    trace_parser.add_argument(
        '-g', '--governor', metavar='value', help='Governor to visualize.'
    )

    market_parser = subparsers.add_parser(
        'market', help='Compare costs in different markets.'
    )
    market_parser.add_argument(
        '-s',
        '--show',
        required=True,
        choices=['load', 'strategy'],
        metavar="['load', 'strategy']"
    )
    market_parser.add_argument(
        '-o',
        '--od_price',
        required=True,
        type=float,
        metavar='float',
        help='On-demand energy price.'
    )
    market_parser.add_argument(
        '-d',
        '--dayahead_prices',
        required=True,
        metavar='path',
        help='Path to day-ahead energy prices (expecting a CSV file).'
    )
    market_parser.add_argument(
        '-i',
        '--imbalance_prices',
        required=True,
        metavar='path',
        help='Path to imbalance energy prices (expecting a CSV file).'
    )

    schedule_parser = subparsers.add_parser(
        'decision', help='Optimize fine-grained decision-making.'
    )
    schedule_parser.add_argument(
        '-o',
        '--option',
        required=True,
        choices=['score', 'schedule'],
        metavar="['score', 'schedule']",
        help=
        "Choose 'score' to compute the agreement accuracy (AA) sore of the predictions; choose 'schedule' for DVFS scheduling."
    )
    schedule_parser.add_argument(
        '-f',
        '--factor',
        type=float,
        metavar='float',
        help='Damping factor of the DVFS scheduler.'
    )
    schedule_parser.add_argument(
        '-d',
        '--dayahead_prices',
        required=True,
        metavar='path',
        help='Path to day-ahead energy prices (expecting a CSV file).'
    )
    schedule_parser.add_argument(
        '-i',
        '--imbalance_prices',
        required=True,
        metavar='path',
        help='Path to imbalance energy prices (expecting a CSV file).'
    )
    schedule_parser.add_argument(
        '-p',
        '--predictions',
        required=True,
        metavar='path',
        help='Machine learning predictions (expecting a CSV file).'
    )
    schedule_parser.add_argument(
        '-a',
        '--aggregate',
        required=True,
        choices=['first', 'last', 'mean'],
        metavar="['first', 'last', 'mean']",
        help="Aggregation method for machine learning predictions."
    )
    schedule_parser.add_argument(
        '-s',
        '--save_to',
        default=None,
        metavar='path',
        help='Destination path of the DVFS schedule.'
    )

    args = parser.parse_args()

    print()
    logging.info('Starting.')
    logging.info('Preprocessing traces.')
    data_trace = pd.read_parquet(args.trace)
    df_trace = preprocess_traces(data_trace, pue=args.pue, map_governor=True)

    return dispatch(args, df_trace)


if __name__ == "__main__":
    sys.exit(main())
